
//////////////////////////////////////////////////////////////////////////////
// Includes

#include "fluids/fluids.h"
#include "simplicialComplex/io/io.h"
#include "simplicialComplex/iterator.h"
#include "simplicialComplex/vertex.h"
#include "simplicialComplex/edge.h"
#include "simplicialComplex/face.h"
#include "simplicialComplex/tet.h"
#include "simplicialComplex/edgeFaceIterator.h"
#include "simplicialComplex/edgeTetIterator.h"
#include "fluids/octree.h"
#include "util/geometry.h"
#include "util/color.h"
#include "util/assert.h"

#include <algorithm>


//////////////////////////////////////////////////////////////////////////////
// Default Constructor

Fluid3D::Fluid3D()
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// Destructor

Fluid3D::~Fluid3D()
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// LoadMesh

bool Fluid3D::LoadMesh( const char *filename, bool freeBoundaries )
{
    if( !Read(m_mesh, filename) )
        return false;

    LOG(info) << "Initializing arrays";
    InitializeArrays();

    LOG(info) << "Building matrices";
    BuildOrLoadMatrices( freeBoundaries, filename );

    LOG(info) << "Building Whitney basis";
    BuildWhitneyBasis();

    LOG(info) << "Building dual cells";
    BuildDualCells();

    LOG(info) << "Placing particles";
    PlaceParticles();

    return true;
}

//////////////////////////////////////////////////////////////////////////////
// InitializeArrays - Resize them and set everything to 0

void Fluid3D::InitializeArrays()
{
    m_velocityField.assign( m_mesh.NumTets(), 0.0f );
    m_velocityFieldBary.assign( m_mesh.NumTets(), 0.0f );
    m_velocityFieldVertices.assign( m_mesh.NumVerts(), 0.0f );
    m_fluxes.Init( m_mesh );
    m_potentials.Init( m_mesh );
    m_vorticities.Init( m_mesh );
    m_forces.Init( m_mesh );
    m_scratchD2.Init( m_mesh );

    m_particleTracker.resize( m_numParticles );

    InitializeVizArrays();
}

//////////////////////////////////////////////////////////////////////////////
// PlaceParticles - Place particles uniformly over the whole domain and keep
//                  track of their barycentric coordinates. Also find and
//                  keep track of the barycentric coordinates of all the tet
//                  circumcenters.

void Fluid3D::PlaceParticles()
{
    Octree octree( &m_mesh );

    vec3 v0, v1, v2, v3;
    for( size_t i = 0; i < m_numParticles; )
    {
        vec3 particle( (real)2.0 * ( rand() / (real)RAND_MAX ) - (real)1.0,
                       (real)2.0 * ( rand() / (real)RAND_MAX ) - (real)1.0,
                       (real)2.0 * ( rand() / (real)RAND_MAX ) - (real)1.0 );

        //// Torus
        //particle *= 3.0f;
        ////particle[1] = particle[1] / 12.0f - 0.5f;
        //particle[1] /= 3.0f;
        //if( particle.square_norm() < (real)1.0 )
        //    continue;

        // Bunny
        if( particle.square_norm() > (real)1.0 )
            continue;
        particle *= 0.15f;
        particle[1] -= 0.25f;

        //// Knot
        //if( particle[0]*particle[0] + particle[2]*particle[2] > (real)1.0 )
        //    continue;
        //particle *= 0.3f;
        //particle[1] *= 0.1f;
        //particle[1] -= 1.4f;


        //// Rings
        //const real r1 = 0.5f;
        //const real r2 = 0.1f;
        //real d = sqrt( particle[0] * particle[0] + particle[2] * particle[2] ) - r1;
        //d *= d;
        //if( d + particle[1] * particle[1] > r2 * r2 )
        //    continue;
        //particle[1] -= 0.5f;

        //// Collision
        //vec3 p1 = particle;
        //vec3 p2 = particle;
        //p1[1] -= (real)0.9;
        //p2[1] += (real)0.9;
        //if( p1.square_norm() > (real)0.01 && p2.square_norm() > (real)0.01 )
        //    continue;


        Tet tet = octree.Locate( particle );
        if( tet )
        {
            m_particleTracker[i].tet = tet;
            tet.Vertices( v0, v1, v2, v3 );
            m_particleTracker[i].baryCoords = BarycentricCoords( particle, v0, v1, v2, v3 );
            m_particles[i++] = vec3f( (float)particle[0], (float)particle[1], (float)particle[2] );

            //real h = atan2( particle[0], particle[2] ) / (2*PI) + (real)0.5;
            //m_colors[i-1] = HueToRGB( h );

            //if( particle[1] < 0 )
            //    m_colors[i-1].set_value( 0, 0, 1 );
            //else
            //    m_colors[i-1].set_value( 1, 0, 0 );
        }
    }


    m_dualVertices.resize( m_mesh.NumTets() );

    vec3 v[4];
    vec4 baryCoords;

    for( SimplexIterator<Tet> tet = m_mesh.Tets(); tet; ++tet )
    {
        vec3 position = tet->Center();
        Tet t = octree.Locate( position );
        if( t )
        {
            t.Vertices( v[0], v[1], v[2], v[3] );
            baryCoords = BarycentricCoords( position, v[0], v[1], v[2], v[3] );
        }
        else
        {
            t = (*tet);
            t.Vertices( v[0], v[1], v[2], v[3] );
            baryCoords = BarycentricCoords( position, v[0], v[1], v[2], v[3] );
            if( baryCoords[0] < 0 ) baryCoords[0] = 0;
            if( baryCoords[1] < 0 ) baryCoords[1] = 0;
            if( baryCoords[2] < 0 ) baryCoords[2] = 0;
            if( baryCoords[3] < 0 ) baryCoords[3] = 0;
            baryCoords /= (baryCoords[0] + baryCoords[1] + baryCoords[2] + baryCoords[3]);
        }

        m_dualVertices[ tet->Index() ].tet = t;
        m_dualVertices[ tet->Index() ].baryCoords = baryCoords;
    }
}

//////////////////////////////////////////////////////////////////////////////
// BuildDualCells - For use with dual 1-form interpolation

void Fluid3D::BuildDualCells()
{
    m_dualCells.resize( m_mesh.NumVerts() );

    // Each tet adds itself to the cells of its 4 vertices
    for( SimplexIterator<Tet> tet = m_mesh.Tets(); tet; ++tet )
    {
        Vertex v[4];
        tet->Vertices( v[0], v[1], v[2], v[3] );
        for( int i = 0; i < 4; ++i )
        {
            Polyhedron::PolyhedronVertex pv;
            pv.position = tet->Center();
            pv.tetID = tet->Index();
            m_dualCells[ v[i].Index() ].m_vertices.push_back( pv );
        }
    }

    std::vector< std::vector<int> > t( m_mesh.NumVerts() );
    for( size_t i = 0; i < t.size(); ++i )
    {
        if( m_dualCells[i].m_vertices.size() > m_baryWeights.size() )
            m_baryWeights.resize( m_dualCells[i].m_vertices.size() );

        t[i].assign( m_dualCells[i].m_vertices.size(), 0 );
    }

    // The edges adjacent to each vertex correspond to the (dual) faces of the
    // dual polyhedron. We represent these dual faces by their plane equations.
    for( SimplexIterator<Edge> edge = m_mesh.Edges(); edge; ++edge )
    {
        Vertex v0, v1;
        edge->Vertices( v0, v1 );
        vec3 normal = v1.Position() - v0.Position();
        normal.normalize();

        size_t v0i = v0.Index();
        size_t v1i = v1.Index();

        m_dualCells[ v0i ].m_planeNormals.push_back(  normal );
        m_dualCells[ v1i ].m_planeNormals.push_back( -normal );

        for( EdgeTetIterator tet = edge->Tets(); tet; ++tet )
        {
            for( size_t j = 0; j < m_dualCells[ v0i ].m_vertices.size(); ++j )
                if( m_dualCells[ v0i ].m_vertices[j].tetID == tet->Index() )
                    m_dualCells[ v0i ].m_vertices[j].planes[t[v0i][j]++] = (int)m_dualCells[ v0i ].m_planeNormals.size() - 1;

            for( size_t j = 0; j < m_dualCells[ v1i ].m_vertices.size(); ++j )
                if( m_dualCells[ v1i ].m_vertices[j].tetID == tet->Index() )
                    m_dualCells[ v1i ].m_vertices[j].planes[t[v1i][j]++] = (int)m_dualCells[ v1i ].m_planeNormals.size() - 1;
        }
    }

    // Precalculate some coefficients that are used to find barycentric coordinates
    for( size_t cell = 0; cell < m_dualCells.size(); ++cell )
        m_dualCells[cell].CalculateCoefficients();

    BuildTetToDualMap();
}

//////////////////////////////////////////////////////////////////////////////
// BuildTetToDualMap - Remember which dual cells overlap with each primal tet

void Fluid3D::BuildTetToDualMap()
{
    m_tetToDualCellMap.resize( m_mesh.NumTets() );

    for( SimplexIterator<Tet> tet = m_mesh.Tets(); tet; ++tet )
    {
        Vertex v[4];
        tet->Vertices( v[0], v[1], v[2], v[3] );

        // The four vertices of the tet are automatically in
        m_tetToDualCellMap.reserve(8);
        m_tetToDualCellMap[ tet->Index() ].push_back( v[0].Index() );
        m_tetToDualCellMap[ tet->Index() ].push_back( v[1].Index() );
        m_tetToDualCellMap[ tet->Index() ].push_back( v[2].Index() );
        m_tetToDualCellMap[ tet->Index() ].push_back( v[3].Index() );

        // Neighbor tets are candidates if their circumcenters are in this tet
        Face f[4];
        tet->Faces( f[0], f[1], f[2], f[3] );
        for( int i = 0; i < 4; ++i )
        {
            Tet neighbor = tet->AdjacentTet( f[i] );
            if( neighbor &&
                IsPointInTet( neighbor.Center(), v[0].Position(), v[1].Position(), v[2].Position(), v[3].Position() ) )
            {
                m_tetToDualCellMap[ tet->Index() ].push_back( neighbor.OppositeVertex( f[i] ).Index() );
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
// BuildWhitneyBasis - For Whitney 2-forms. Converts fluxes to piecewise-
//                     constant velocities defined on tets

void Fluid3D::BuildWhitneyBasis()
{
    m_whitneyBasis.resize( m_mesh.NumTets() );
    m_fluxesToBaryVecs.resize( m_mesh.NumTets() );

    Face f0, f1, f2, f3;
    vec3 grad0, grad1, grad2, grad3;
    real h0, h1, h2, h3;

    for( SimplexIterator<Tet> tet = m_mesh.Tets(); tet; ++tet )
    {
        tet->Faces( f0, f1, f2, f3 );

        // gradX is a vector that points towards vertex X, is orthogonal to the opposite face,
        // and has length equal to the 1/height of the tet
        grad0 = f0.Normal();
        h0 = grad0.dot( tet->OppositeVertex(f0).Position() - f0.Center() );
        grad0 /= h0;

        grad1 = f1.Normal();
        h1 = grad1.dot( tet->OppositeVertex(f1).Position() - f1.Center() );
        grad1 /= h1;

        grad2 = f2.Normal();
        h2 = grad2.dot( tet->OppositeVertex(f2).Position() - f2.Center() );
        grad2 /= h2;

        grad3 = f3.Normal();
        h3 = grad3.dot( tet->OppositeVertex(f3).Position() - f3.Center() );
        grad3 /= h3;

        // Whitney 2-form basis
        // The sign takes into account whether positive flux means entering or exiting for that particular face
        m_whitneyBasis[ tet->Index() ][0] = ( grad2.cross(grad1) + grad1.cross(grad3) + grad3.cross(grad2) ) * ( (h0 < 0) ? (real)(-0.5) : (real)0.5 );
        m_whitneyBasis[ tet->Index() ][1] = ( grad3.cross(grad0) + grad0.cross(grad2) + grad2.cross(grad3) ) * ( (h1 < 0) ? (real)(-0.5) : (real)0.5 );
        m_whitneyBasis[ tet->Index() ][2] = ( grad0.cross(grad3) + grad3.cross(grad1) + grad1.cross(grad0) ) * ( (h2 < 0) ? (real)(-0.5) : (real)0.5 );
        m_whitneyBasis[ tet->Index() ][3] = ( grad1.cross(grad2) + grad2.cross(grad0) + grad0.cross(grad1) ) * ( (h3 < 0) ? (real)(-0.5) : (real)0.5 );

        real scale = (real)(-1.0) / ( (real)3.0 * tet->Volume() );
        m_fluxesToBaryVecs[ tet->Index() ][0] = scale * ( (h0 < 0) ? 1 : -1 );
        m_fluxesToBaryVecs[ tet->Index() ][1] = scale * ( (h1 < 0) ? 1 : -1 );
        m_fluxesToBaryVecs[ tet->Index() ][2] = scale * ( (h2 < 0) ? 1 : -1 );
        m_fluxesToBaryVecs[ tet->Index() ][3] = scale * ( (h3 < 0) ? 1 : -1 );
    }
}

//////////////////////////////////////////////////////////////////////////////
// AdvectParticles - Trace them forwards through the velocity field

void Fluid3D::AdvectParticles()
{
    for( size_t curParticle = 0; curParticle < m_numParticles; ++curParticle )
    {
        // Trace a path forward through the velocity field for time dt.
        PathTrace( false, m_particleTracker[curParticle], true );
        //PathTraceSmooth( false, m_particleTracker[curParticle], true );

#ifdef REAL_IS_FLOAT
        m_particles[curParticle] = m_particleTracker[curParticle].GetPosition();
#else
        vec3 pos = m_particleTracker[curParticle].GetPosition();
        m_particles[curParticle].set_value( (float)pos[0], (float)pos[1], (float)pos[2] );
#endif

        //float h = 3.0f * GetCirculation( m_particleTracker[curParticle] ).length();
        //h = 0.75f * (1 - h);
        //m_colors[curParticle] = HueToRGB( h );
    }
}

//////////////////////////////////////////////////////////////////////////////
// AdvectVorticities - by backtracking the dual loops and integrating

void Fluid3D::AdvectVorticities()
{
    // Advect the vertices of the dual cells
    std::vector<vec3> advectedDualVertices( m_dualVertices.size() );
    std::vector<vec3> sampledField( m_dualVertices.size() );
    for( size_t i = 0; i < m_dualVertices.size(); ++i )
    {
        MeshPoint3 advectedDualVertex = m_dualVertices[i];
        PathTrace( true, advectedDualVertex, false );
        advectedDualVertices[i] = advectedDualVertex.GetPosition();
        sampledField[i] = GetCirculation( advectedDualVertex );
    }

    Form<DUAL,2> & advectedVorticities = m_scratchD2;
    advectedVorticities.Set( 0.0 );

    Tet t0, t1;
    for( SimplexIterator<Edge> edge = m_mesh.Edges(); edge; ++edge )
    {
        if( !edge->IsOnBoundary() )
        {
            for( EdgeFaceIterator face = edge->Faces(); face; ++face )
            {
                face->Tets( t0, t1 );
                vec3 pieceCirculation = (real)0.5 * ( sampledField[ t0.Index() ] + sampledField[ t1.Index() ] );
                advectedVorticities[ edge->Index() ] += face.RelativeSign() * pieceCirculation.dot( advectedDualVertices[ t0.Index() ] - advectedDualVertices[ t1.Index() ] );
            }
        }
    }

    m_vorticities.Swap( advectedVorticities );
}

//////////////////////////////////////////////////////////////////////////////
// AddForces

void Fluid3D::AddForces()
{
    m_vorticities.Add( m_forces );

    // Forces due to buoyancy
    if( m_particleBouyancy > 0 )
    {
        m_fluxes.Set( 0.0 );

        Vector densities( m_mesh.NumTets() );
        densities.Set( 0.0f );
        for( size_t i = 0; i < m_particleTracker.size(); ++i )
            densities[ m_particleTracker[i].tet.Index() ] += m_particleBouyancy / m_particleTracker[i].tet.Volume();

        Tet t0, t1;
        for( SimplexIterator<Face> face = m_mesh.Faces(); face; ++face )
        {
            if( !face->IsOnBoundary() )
            {
                face->Tets( t0, t1 );
                m_fluxes[face->Index()] += m_timestep * face->Area() * face->Normal()[1] * ( densities[t0.Index()] + densities[t1.Index()] );
            }
        }

        // Add the forces into the current field via its vorticity
        Form<DUAL,2> & forceVorticities = m_scratchD2;
        d_star_P2.Apply( m_fluxes, forceVorticities );
        m_vorticities.Add( forceVorticities );
    }
}

//////////////////////////////////////////////////////////////////////////////
// GetCirculation - Finds the interpolated dual velocity at this point

vec3 Fluid3D::GetCirculation( const MeshPoint3& point )
{
    vec3 position = point.GetPosition();

    const std::vector<size_t> & dualCellsToCheck = m_tetToDualCellMap[ point.tet.Index() ];

    // We know which voronoi cell we're in by finding the closest vertex
    size_t closestVertex = size_t(-1);
    real smallest = max_real();
    for( size_t i = 0; i < dualCellsToCheck.size(); ++i )
    {
        real distance = ( position - m_mesh.GetVertex( dualCellsToCheck[i] ).Position() ).square_norm();
        if( distance < smallest )
        {
            closestVertex = dualCellsToCheck[i];
            smallest = distance;
        }
    }
    smart_assert( closestVertex != size_t(-1) ).fatal();

    //if( m_mesh.GetVertex(closestVertex).IsOnBoundary() )
    //    return m_velocityField[ point.tet.Index() ];

    // Find the barycentric coordinates
    m_dualCells[ closestVertex ].BarycentricCoords( position, m_baryWeights );

    // And do the interpolation
    vec3 retval(0,0,0);
    for( size_t i = 0; i < m_dualCells[ closestVertex ].m_vertices.size(); ++i )
        retval += m_baryWeights[i] * m_velocityField[ m_dualCells[ closestVertex ].m_vertices[i].tetID ];

    return retval;
}

//////////////////////////////////////////////////////////////////////////////
// PermuteBaryCoords - this is used when stepping over a face from one tet to
//                     another. The barycentric coordinates of the point on
//                     the face are converted from the basis of the first tet
//                     to the basis of the second tet. PrevTet and NextTet
//                     should have exactly three integers in common.

static vec4 PermuteBaryCoords( const vec4& original, const Tet& prevTet, const Tet& nextTet )
{
    int iPrevTet[4], iNextTet[4];
    prevTet.Vertices( iPrevTet[0], iPrevTet[1], iPrevTet[2], iPrevTet[3] );
    nextTet.Vertices( iNextTet[0], iNextTet[1], iNextTet[2], iNextTet[3] );

    vec4 permuted(0,0,0,0);

    // unrolled the loop
    if     ( iNextTet[0] == iPrevTet[0] ) permuted[0] = original[0];
    else if( iNextTet[0] == iPrevTet[1] ) permuted[0] = original[1];
    else if( iNextTet[0] == iPrevTet[2] ) permuted[0] = original[2];
    else if( iNextTet[0] == iPrevTet[3] ) permuted[0] = original[3];

    if     ( iNextTet[1] == iPrevTet[0] ) permuted[1] = original[0];
    else if( iNextTet[1] == iPrevTet[1] ) permuted[1] = original[1];
    else if( iNextTet[1] == iPrevTet[2] ) permuted[1] = original[2];
    else if( iNextTet[1] == iPrevTet[3] ) permuted[1] = original[3];

    if     ( iNextTet[2] == iPrevTet[0] ) permuted[2] = original[0];
    else if( iNextTet[2] == iPrevTet[1] ) permuted[2] = original[1];
    else if( iNextTet[2] == iPrevTet[2] ) permuted[2] = original[2];
    else if( iNextTet[2] == iPrevTet[3] ) permuted[2] = original[3];

    if     ( iNextTet[3] == iPrevTet[0] ) permuted[3] = original[0];
    else if( iNextTet[3] == iPrevTet[1] ) permuted[3] = original[1];
    else if( iNextTet[3] == iPrevTet[2] ) permuted[3] = original[2];
    else if( iNextTet[3] == iPrevTet[3] ) permuted[3] = original[3];

    return permuted;
}

//////////////////////////////////////////////////////////////////////////////
// PathTrace - trace a path through the complex, one tet at a time.

void Fluid3D::PathTrace( bool negativeDirection, MeshPoint3& point, bool updateStreamlines )
{
    // Walk along in the direction of (+/-)velocity until we've exhausted
    // 'dt' worth of movement, or until we've hit the boundary, whichever comes first.
    // If we hit a face, then walk over to the next tet and repeat.

    Tet& tet = point.tet;
    vec4& posBary = point.baryCoords;

    Tet nextTet = tet;
    vec4 dirBary;

    real timeToIntersection;
    real dt = m_timestep;
    const real sign = negativeDirection ? (real)(-1.0) : (real)1.0;

    if( m_numStreamlines == 0 || m_numPointsPerStreamline == 0 )
        updateStreamlines = false;

    if( updateStreamlines )
    {
        BeginStreamline();
        AddStreamlinePoint( point.GetPosition() );
    }

    // while we haven't hit a boundary
    while( nextTet )
    {
        // put the barycentric coordinates into the frame of reference of the next tet
        posBary = PermuteBaryCoords( posBary, tet, nextTet );
        tet = nextTet;

        dirBary = sign * m_velocityFieldBary[ tet.Index() ];

        // note that since we're passing it an unnormalized direction vector,
        // the distance returned will be scaled by 1 / vector length.
        // And since the direction vector is a velocity vector (which has units
        // of distance-per-unit-time) then the distance returned is actually time.
        timeToIntersection = TraceTet( tet, posBary, dirBary, nextTet );

        // if the nearest intersection is too far
        if( timeToIntersection > dt )
        {
            // back up to where we ran out of time
            timeToIntersection = dt;
            // and indicate that we're done
            nextTet = Tet();
        }

        // step to the intersection point
        posBary += timeToIntersection * dirBary;
        dt -= timeToIntersection;

        if( updateStreamlines )
            AddStreamlinePoint( point.GetPosition() );
    }

    if( updateStreamlines )
        EndStreamline();
}

//////////////////////////////////////////////////////////////////////////////
// PathTraceSmooth - RK2 version

void Fluid3D::PathTraceSmooth( bool negativeDirection, MeshPoint3& point, bool updateStreamlines )
{
    // Walk along in the direction of (+/-)velocity until we've exhausted
    // 'dt' worth of movement, or until we've hit the boundary, whichever comes first.
    // If we hit a face, then walk over to the next tet and repeat.

    Tet& tet = point.tet;
    vec4& posBary = point.baryCoords;

    Tet nextTet = tet;
    vec4 dirBary;

    real timeToIntersection;
    real dt = m_timestep;
    const real sign = negativeDirection ? (real)(-1.0) : (real)1.0;

    if( m_numStreamlines == 0 || m_numPointsPerStreamline == 0 )
        updateStreamlines = false;

    if( updateStreamlines )
    {
        BeginStreamline();
        AddStreamlinePoint( point.GetPosition() );
    }

    // while we haven't hit a boundary
    while( nextTet )
    {
        // put the barycentric coordinates into the frame of reference of the next tet
        posBary = PermuteBaryCoords( posBary, tet, nextTet );
        tet = nextTet;

        vec3 v[4];
        tet.Vertices( v[0], v[1], v[2], v[3] );

        int vi[4];
        tet.Vertices( vi[0], vi[1], vi[2], vi[3] );

        //vec3 dir = sign * GetCirculation( point );
        vec3 dir = sign * Affine( point.baryCoords, m_velocityFieldVertices[ vi[0] ],
                                                    m_velocityFieldVertices[ vi[1] ],
                                                    m_velocityFieldVertices[ vi[2] ],
                                                    m_velocityFieldVertices[ vi[3] ] );

        //if( tet.IsOnBoundary() )
        //    dir = sign * m_velocityField[ tet.Index() ];

        dirBary = BarycentricCoords( v[0] + dir, v[0], v[1], v[2], v[3] ) - vec4(1,0,0,0);
        //dirBary = sign * m_velocityFieldBary[ tet.Index() ];

        if( (posBary[0] <= 0 && dirBary[0] < 0) ||
            (posBary[1] <= 0 && dirBary[1] < 0) ||
            (posBary[2] <= 0 && dirBary[2] < 0) ||
            (posBary[3] <= 0 && dirBary[3] < 0) )
        {
            if( posBary[0] <= 0 ) posBary[0] = 1e-1f;
            if( posBary[1] <= 0 ) posBary[1] = 1e-1f;
            if( posBary[2] <= 0 ) posBary[2] = 1e-1f;
            if( posBary[3] <= 0 ) posBary[3] = 1e-1f;
            posBary /= posBary[0] + posBary[1] + posBary[2] + posBary[3];
        }

        // note that since we're passing it an unnormalized direction vector,
        // the distance returned will be scaled by 1 / vector length.
        // And since the direction vector is a velocity vector (which has units
        // of distance-per-unit-time) then the distance returned is actually time.
        timeToIntersection = TraceTet( tet, posBary, dirBary, nextTet );

        // if the nearest intersection is too far
        if( timeToIntersection > dt )
        {
            // back up to where we ran out of time
            timeToIntersection = dt;
            // and indicate that we're done
            nextTet = Tet();
        }

        {
            //// RK2
            //dir = sign * Affine( posBary + (real)0.5 * timeToIntersection * dirBary,
            //                     m_velocityFieldVertices[ vi[0] ],
            //                     m_velocityFieldVertices[ vi[1] ],
            //                     m_velocityFieldVertices[ vi[2] ],
            //                     m_velocityFieldVertices[ vi[3] ] );
            //dirBary = BarycentricCoords( v[0] + dir, v[0], v[1], v[2], v[3] ) - vec4(1,0,0,0);

            //if( (posBary[0] == 0 && dirBary[0] < 0) ||
            //    (posBary[1] == 0 && dirBary[1] < 0) ||
            //    (posBary[2] == 0 && dirBary[2] < 0) ||
            //    (posBary[3] == 0 && dirBary[3] < 0) )
            //{
            //    if( posBary[0] == 0 ) posBary[0] = 1e-1f;
            //    if( posBary[1] == 0 ) posBary[1] = 1e-1f;
            //    if( posBary[2] == 0 ) posBary[2] = 1e-1f;
            //    if( posBary[3] == 0 ) posBary[3] = 1e-1f;
            //    posBary /= posBary[0] + posBary[1] + posBary[2] + posBary[3];
            //}

            //timeToIntersection = TraceTet( tet, posBary, dirBary, nextTet );

            //if( timeToIntersection > dt )
            //{
            //    timeToIntersection = dt;
            //    nextTet = Tet();
            //}
        }

        // step to the intersection point
        posBary += timeToIntersection * dirBary;
        dt -= timeToIntersection;

        if( updateStreamlines )
            AddStreamlinePoint( point.GetPosition() );
    }

    if( updateStreamlines )
        EndStreamline();
}

//////////////////////////////////////////////////////////////////////////////
// TraceTet - trace a ray through one tet of the mesh. Finds the point
//            where the ray exits the current tet (closest face), and gives
//            the intersection distance as well as the tet on the other side
//            of that face.

real Fluid3D::TraceTet( const Tet& tet, const vec4& posBary, const vec4& dirBary, Tet& nextTet )
{
    real t0 = -posBary[0] / dirBary[0];
    real t1 = -posBary[1] / dirBary[1];
    real t2 = -posBary[2] / dirBary[2];
    real t3 = -posBary[3] / dirBary[3];

    // Set the intersections that are negative (behind the ray origin) to
    // a big positive number to make the comparisons easier.
    // In the assignments above, division by zero is gracefully handled because
    // it sets the distance to infinity. However, if the numerator is also zero
    // we get NaN. To deal with that I use a negated greater-than rather than a
    // less-or-equal.
    if( !(t0 > 0) )  t0 = max_real();
    if( !(t1 > 0) )  t1 = max_real();
    if( !(t2 > 0) )  t2 = max_real();
    if( !(t3 > 0) )  t3 = max_real();

    real closest = t0;
    int whichNeighbor = 0;
    if( t1 < closest )
    {
        closest = t1;
        whichNeighbor = 1;
    }
    if( t2 < closest )
    {
        closest = t2;
        whichNeighbor = 2;
    }
    if( t3 < closest )
    {
        closest = t3;
        whichNeighbor = 3;
    }

    nextTet = tet.AdjacentTet( whichNeighbor );

    return closest;
}

//////////////////////////////////////////////////////////////////////////////
// FluxesToVorticities - Just hits m_fluxes with a d*

void Fluid3D::FluxesToVorticities()
{
    d_star_P2.Apply( m_fluxes, m_vorticities );
}

//////////////////////////////////////////////////////////////////////////////
// VorticitiesToFluxes - Solve for fluxes given vorticites and divergence-
//                       free constraint

void Fluid3D::VorticitiesToFluxes()
{
    // Need to find fluxes such that d*(fluxes) = vorticities,
    // with constraint d(fluxes) = 0. This can be found by solving
    // d*d(potentials) = vorticities, fluxes = d(potentials).
    // Note that *d*d*(potentials) = 0; that term is there because
    // d*d by itself is singular.
    sdsdspdsd_P1.ApplyInverse( m_vorticities, m_potentials );
    d_P1.Apply( m_potentials, m_fluxes );
}

//////////////////////////////////////////////////////////////////////////////
// FluxesToVelocities - Converts the flux field into a piecewise-constant
//                      velocity field defined on tets, using Whitney forms

void Fluid3D::FluxesToVelocities()
{
    Face f0, f1, f2, f3;
    for( SimplexIterator<Tet> tet = m_mesh.Tets(); tet; ++tet )
    {
        tet->Faces( f0, f1, f2, f3 );

        m_velocityFieldBary[ tet->Index() ].set_value(
            (real)m_fluxes[ f0.Index() ] * m_fluxesToBaryVecs[ tet->Index() ][0],
            (real)m_fluxes[ f1.Index() ] * m_fluxesToBaryVecs[ tet->Index() ][1],
            (real)m_fluxes[ f2.Index() ] * m_fluxesToBaryVecs[ tet->Index() ][2],
            (real)m_fluxes[ f3.Index() ] * m_fluxesToBaryVecs[ tet->Index() ][3] );

        m_velocityField[ tet->Index() ] =
            (real)m_fluxes[ f0.Index() ] * m_whitneyBasis[ tet->Index() ][0] +
            (real)m_fluxes[ f1.Index() ] * m_whitneyBasis[ tet->Index() ][1] +
            (real)m_fluxes[ f2.Index() ] * m_whitneyBasis[ tet->Index() ][2] +
            (real)m_fluxes[ f3.Index() ] * m_whitneyBasis[ tet->Index() ][3];

    }

    for( SimplexIterator<Vertex> vertex = m_mesh.Vertices(); vertex; ++vertex )
    {
        vec3 position = vertex->Position();
        size_t v = vertex->Index();
        // barycentric interpolation on the dual
        m_dualCells[v].BarycentricCoords( position, m_baryWeights );
        m_velocityFieldVertices[v].set_value(0,0,0);
        for( size_t i = 0; i < m_dualCells[v].m_vertices.size(); ++i )
            m_velocityFieldVertices[v] += m_baryWeights[i] * m_velocityField[ m_dualCells[v].m_vertices[i].tetID ];
    }
}
