//////////////////////////////////////////////////////////////////////////////
// LinAlgImpl.h
// Private include for Vector and Sparse Matrix implementations.
//////////////////////////////////////////////////////////////////////////////

#ifndef _LIN_ALG_IMPL_H_
#define _LIN_ALG_IMPL_H_


//////////////////////////////////////////////////////////////////////////////
// Includes

#ifdef COMPILE_WITH_PETSC
#include "petscksp.h"
#endif

#include <vector>
#include <map>


//////////////////////////////////////////////////////////////////////////////
// class VectorImpl
// Abstract base class for Vector implementations

class VectorImpl
{
public:
    // Same public interface as Vector
    virtual ~VectorImpl() {};

    virtual       double & operator [] ( size_t i )       = 0;
    virtual const double & operator [] ( size_t i ) const = 0;

    virtual void Assign( size_t size, double value ) = 0;
    virtual void Set   ( double value )              = 0;
    virtual void Copy  ( const VectorImpl *v )       = 0;
    virtual void Swap  ( VectorImpl *v )             = 0;

    virtual void Add       ( const VectorImpl *v )           = 0;
    virtual void Subtract  ( const VectorImpl *v )           = 0;
    virtual void Multiply  ( const VectorImpl *v )           = 0;
    virtual void Multiply  ( double f )                      = 0;
    virtual void Divide    ( const VectorImpl *v )           = 0;
    virtual void Negate    ( void )                          = 0;
    virtual void MAD       ( double f, const VectorImpl *v ) = 0;
    virtual void MAD       ( const VectorImpl *v, double f ) = 0;
    virtual real DotProduct( const VectorImpl *v ) const     = 0;
};

//////////////////////////////////////////////////////////////////////////////
// class MatImpl
// Abstract base class for SparseMatrix implementations

class MatImpl
{
public:
    // Same public interface as SparseMatrix
    virtual ~MatImpl() {}
    virtual void BeginConstruction( int numRows, int numColumns, bool bSymmetric ) = 0;
    virtual void EndConstruction( PRECOND_TYPE_ENUM type, real dropTolerance ) = 0;
    virtual void Save( const char *filename ) = 0;
    virtual void Load( const char *filename ) = 0;
    virtual void Set( int i, int j, real value ) = 0;
    virtual void Add( int i, int j, real value ) = 0;
    virtual void Multiply( const VectorImpl *vec, VectorImpl *product ) const = 0;
    virtual bool Solve( const VectorImpl *b, VectorImpl *x, int maxIterations, real epsilon ) const = 0;

public:
    void LogConvergenceStats( bool enable ) { m_bLogConvergenceStats = enable; }
protected:
    bool m_bLogConvergenceStats;
};


//////////////////////////////////////////////////////////////////////////////
// class MyMatImpl

class MyMatImpl : public MatImpl
{
public:
    MyMatImpl();
    MyMatImpl( int numRows, int numColumns );
    ~MyMatImpl();
    void BeginConstruction( int numRows, int numColumns, bool bSymmetric );
    void EndConstruction( PRECOND_TYPE_ENUM type, real dropTolerance );
    void Save( const char *filename );
    void Load( const char *filename );
    void Set( int i, int j, real value );
    void Add( int i, int j, real value );
    void Multiply( const VectorImpl *vec, VectorImpl *product ) const;
    bool Solve( const VectorImpl *b, VectorImpl *x, int maxIterations, real epsilon ) const;

private:
    void Clear();
    void MAD( const VectorImpl *vec, const VectorImpl *plus, VectorImpl *product ) const;

    class Preconditioner;
    class IdentityPreconditioner;
    class DiagonalPreconditioner;
    class IncompleteCholeskyPreconditioner;

    int  m_numRows;
    int  m_numColumns;
    bool m_bSymmetric;

    // This representation is used during the construction phase
    std::vector< std::map<int, double> > m_columns;

    // A compressed column representation
    Preconditioner *m_preconditioner;
    double *m_values;
    int    *m_colPtr;
    int    *m_rowInd;
};

//////////////////////////////////////////////////////////////////////////////
// class MyVectorImpl

class MyVectorImpl : public VectorImpl
{
    friend class MyMatImpl;
    friend class MyMatImpl::IncompleteCholeskyPreconditioner;

public:
    MyVectorImpl( size_t size );
    ~MyVectorImpl();

          double & operator [] ( size_t i );
    const double & operator [] ( size_t i ) const;

    void Assign( size_t size, double value );
    void Set   ( double value );
    void Copy  ( const VectorImpl *v );
    void Swap  ( VectorImpl *v );

    void Add       ( const VectorImpl *v );
    void Subtract  ( const VectorImpl *v );
    void Multiply  ( const VectorImpl *v );
    void Multiply  ( double f );
    void Divide    ( const VectorImpl *v );
    void Negate    ( void );
    void MAD       ( double f, const VectorImpl *v );
    void MAD       ( const VectorImpl *v, double f );
    real DotProduct( const VectorImpl *v ) const;

private:
    std::vector<double> data;
};


#ifdef COMPILE_WITH_PETSC

//////////////////////////////////////////////////////////////////////////////
// class PetscVectorImpl

class PetscVectorImpl : public VectorImpl
{
    friend class PetscMatImpl;

public:
    PetscVectorImpl( size_t size );
    ~PetscVectorImpl();

          double & operator [] ( size_t i );
    const double & operator [] ( size_t i ) const;

    void Assign( size_t size, double value );
    void Set   ( double value );
    void Copy  ( const VectorImpl *v );
    void Swap  ( VectorImpl *v );

    void Add       ( const VectorImpl *v );
    void Subtract  ( const VectorImpl *v );
    void Multiply  ( const VectorImpl *v );
    void Multiply  ( double f );
    void Divide    ( const VectorImpl *v );
    void Negate    ( void );
    void MAD       ( double f, const VectorImpl *v );
    void MAD       ( const VectorImpl *v, double f );
    real DotProduct( const VectorImpl *v ) const;

private:
    Vec m_PetscVec;
    mutable double *data;
};

//////////////////////////////////////////////////////////////////////////////
// class PetscMatImpl

class PetscMatImpl : public MatImpl
{
public:
    PetscMatImpl();
    PetscMatImpl( int numRows, int numColumns );
    ~PetscMatImpl();
    void BeginConstruction( int numRows, int numColumns, bool bSymmetric );
    void EndConstruction( PRECOND_TYPE_ENUM type, real dropTolerance );
    void Save( const char *filename );
    void Load( const char *filename );
    void Set( int i, int j, real value );
    void Add( int i, int j, real value );
    void Multiply( const VectorImpl *vec, VectorImpl *product ) const;
    bool Solve( const VectorImpl *b, VectorImpl *x, int maxIterations, real epsilon ) const;

private:
    void Clear();
    void InitKSP();

    int  m_numRows;
    int  m_numColumns;
    bool m_bSymmetric;

    // This representation is used during the construction phase
    std::vector< std::map<int, double> > m_rows;

    // The Petsc representation
    KSP m_PetscKSP;
    Mat m_PetscMat;
};

#endif  // COMPILE_WITH_PETSC


#endif  // _LIN_ALG_IMPL_H_
