//////////////////////////////////////////////////////////////////////////////
// LinAlgMyImpl.cpp
// My implementation of the Vector and Sparse Matrix classes.
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Includes

#include "linearAlgebra/linearAlgebra.h"
#include "linearAlgebra/linAlgImpl.h"
#include "util/assert.h"

#include <algorithm>
#include <functional>
#include <numeric>
#include <fstream>

using namespace std;


//////////////////////////////////////////////////////////////////////////////
// MyVectorImpl

MyVectorImpl::MyVectorImpl( size_t size ) : data(size)
{
    // empty
}
MyVectorImpl::~MyVectorImpl()
{
    // empty
}
double & MyVectorImpl::operator [] ( size_t i )
{
    return data[i];
}
const double & MyVectorImpl::operator [] ( size_t i ) const
{
    return data[i];
}
void MyVectorImpl::Assign( size_t size, double value )
{
    data.assign( size, value );
}
void MyVectorImpl::Set( double value )
{
    fill( data.begin(), data.end(), value );
}
void MyVectorImpl::Copy( const VectorImpl *v )
{
    copy( static_cast<const MyVectorImpl*>(v)->data.begin(), static_cast<const MyVectorImpl*>(v)->data.end(), data.begin() );
}
void MyVectorImpl::Swap( VectorImpl *v )
{
    data.swap( static_cast<MyVectorImpl*>(v)->data );
}
void MyVectorImpl::Add( const VectorImpl *v )
{
    transform( data.begin(), data.end(), static_cast<const MyVectorImpl*>(v)->data.begin(), data.begin(), plus<double>() );
}
void MyVectorImpl::Subtract( const VectorImpl *v )
{
    transform( data.begin(), data.end(), static_cast<const MyVectorImpl*>(v)->data.begin(), data.begin(), minus<double>() );
}
void MyVectorImpl::Multiply( const VectorImpl *v )
{
    transform( data.begin(), data.end(), static_cast<const MyVectorImpl*>(v)->data.begin(), data.begin(), multiplies<double>() );
}
void MyVectorImpl::Multiply( double f )
{
    transform( data.begin(), data.end(), data.begin(), bind2nd( multiplies<double>(), f ) );
}
void MyVectorImpl::Divide( const VectorImpl *v )
{
    transform( data.begin(), data.end(), static_cast<const MyVectorImpl*>(v)->data.begin(), data.begin(), divides<double>() );
}
void MyVectorImpl::Negate( void )
{
    transform( data.begin(), data.end(), data.begin(), negate<double>() );
}
struct MADfunc
{
    MADfunc( double _f ) : f(_f) {}
    double operator()( const double& a, const double& b ) { return f * a + b; }
    double f;
};
void MyVectorImpl::MAD( double f, const VectorImpl *v )
{
    transform( data.begin(), data.end(), static_cast<const MyVectorImpl*>(v)->data.begin(), data.begin(), MADfunc(f) );
}
void MyVectorImpl::MAD( const VectorImpl *v, double f )
{
    transform( static_cast<const MyVectorImpl*>(v)->data.begin(), static_cast<const MyVectorImpl*>(v)->data.end(), data.begin(), data.begin(), MADfunc(f) );
}
real MyVectorImpl::DotProduct( const VectorImpl *v ) const
{
    return (real)inner_product( data.begin(), data.end(), static_cast<const MyVectorImpl*>(v)->data.begin(), 0.0 );
}


//////////////////////////////////////////////////////////////////////////////
// class Preconditioner
// Abstract base class for CG preconditioners

class MyMatImpl::Preconditioner
{
public:
    Preconditioner() {}
    virtual ~Preconditioner() {}

    virtual void Apply( const VectorImpl *, VectorImpl * ) = 0;
};

//////////////////////////////////////////////////////////////////////////////
// class IdentityPreconditioner
// Inherits from Preconditioner. Does nothing.

class MyMatImpl::IdentityPreconditioner : public MyMatImpl::Preconditioner
{
public:
    IdentityPreconditioner() {}
    ~IdentityPreconditioner() {}

    void Apply( const VectorImpl *in, VectorImpl *out )
    {
        out->Copy( in );
    }
};

//////////////////////////////////////////////////////////////////////////////
// class DiagonalPreconditioner
// Inherits from Preconditioner

class MyMatImpl::DiagonalPreconditioner : public MyMatImpl::Preconditioner
{
public:
    DiagonalPreconditioner( const MyMatImpl &M );
    ~DiagonalPreconditioner() {}

    void Apply( const VectorImpl *in, VectorImpl *out )
    {
        out->Copy( in );
        out->Multiply( &d );
    }

private:
    MyVectorImpl d;
};

//////////////////////////////////////////////////////////////////////////////
// class IncompleteCholeskyPreconditioner
// Inherits from Preconditioner

class MyMatImpl::IncompleteCholeskyPreconditioner : public MyMatImpl::Preconditioner
{
public:
    IncompleteCholeskyPreconditioner( const MyMatImpl &M, real dropTolerance );
    ~IncompleteCholeskyPreconditioner()
    {
        delete [] diag;
        delete [] offDiag;
        delete [] colPtr;
        delete [] rowInd;
    }

    void Apply( const VectorImpl *in, VectorImpl *out )
    {
        out->Copy( in );

        // forward substitution on L
        int j = 0;
        for( int i = 0; i < N; ++i )
        {
            ((MyVectorImpl*)out)->data[i] *= diag[i];
            //(*out)[i] *= diag[i];
            for( ; j < colPtr[i+1]; ++j )
                ((MyVectorImpl*)out)->data[rowInd[j]] -= (*out)[i] * offDiag[j];
                //(*out)[rowInd[j]] -= (*out)[i] * offDiag[j];
        }
        // back substitution on L^t
        j = colPtr[N]-1;
        for( int i = N-1; i >= 0; --i )
        {
            for( ; j >= colPtr[i]; --j )
                ((MyVectorImpl*)out)->data[i] -= ((MyVectorImpl*)out)->data[rowInd[j]] * offDiag[j];
                //(*out)[i] -= (*out)[rowInd[j]] * offDiag[j];
            ((MyVectorImpl*)out)->data[i] *= diag[i];
            //(*out)[i] *= diag[i];
        }
    }

private:
    int N;
    double *diag;     // stores (1 / the diagonal entry)
    double *offDiag;
    int *colPtr;
    int *rowInd;
};


//////////////////////////////////////////////////////////////////////////////
// DiagonalPreconditioner Constructor - Creates a diagonal matrix that
//                                      contains the inverse of the diagonal
//                                      elements of M

MyMatImpl::DiagonalPreconditioner::DiagonalPreconditioner( const MyMatImpl &M ) : d(M.m_numRows)
{
    for( int i = 0; i < M.m_numRows; ++i )
    {
        d[i] = 1.0f / M.m_values[ M.m_colPtr[i] ];
    }
}

//////////////////////////////////////////////////////////////////////////////
// IncompleteCholeskyPreconditioner Constructor - Computes and stores the
//                                                factorization

MyMatImpl::IncompleteCholeskyPreconditioner::IncompleteCholeskyPreconditioner( const MyMatImpl &M, real dropTolerance )
{
    smart_assert( M.m_bSymmetric ).msg( "Solver requires a symmetric matrix" ).fatal();

    N = M.m_numRows;
    diag = new double[N];
    vector< map<int,double> > C(N);

    // put the diagonals in diag, and the off-diagonals in C[i]
    int j = 0;
    for( int i = 0; i < N; ++i )
    {
        diag[i] = M.m_values[j++];
        for( ; j < M.m_colPtr[i+1]; ++j )
            C[i].insert( pair<int,double>( M.m_rowInd[j], M.m_values[j] ) );
    }

    // factor
    for( int i = 0; i < N; ++i )
    {
        diag[i] = sqrt( diag[i] );
        for( map<int,double>::iterator it1 = C[i].begin(); it1 != C[i].end(); )
        {
            it1->second /= diag[i];

            if( abs(it1->second) < dropTolerance )
            {
                C[i].erase(it1++);
                continue;
            }

            diag[it1->first] -= it1->second * it1->second;
            for( map<int,double>::iterator it2 = C[i].begin(); it2 != it1; ++it2 )
                C[it2->first][it1->first] -= it2->second * it1->second;

            ++it1;
        }
    }

    // convert to compressed representation
    colPtr = new int[N+1];
    int size = 0;
    for( int i = 0; i < N; ++i )
    {
        colPtr[i] = size;
        size += (int)C[i].size();
    }
    colPtr[N] = size;

    offDiag = new double[size];
    rowInd = new int[size];
    j = 0;
    for( int i = 0; i < N; ++i )
    {
        for( map<int,double>::iterator it = C[i].begin(); it != C[i].end(); ++it, ++j )
        {
            rowInd[j] = it->first;
            offDiag[j] = it->second;
        }
        // invert the entries in diag, so that later we can multiply instead of divide
        diag[i] = 1 / diag[i];
    }
}


//////////////////////////////////////////////////////////////////////////////
// Default Constructor

MyMatImpl::MyMatImpl()
    : m_numRows(0),
      m_numColumns(0),
      m_bSymmetric(false),
      m_preconditioner(0),
      m_values(0),
      m_colPtr(0),
      m_rowInd(0)
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// Constructor - takes matrix size as arguments

MyMatImpl::MyMatImpl( int numRows, int numColumns )
    : m_numRows(numRows),
      m_numColumns(numColumns),
      m_bSymmetric(false),
      m_columns(numColumns),
      m_preconditioner(0),
      m_values(0),
      m_colPtr(0),
      m_rowInd(0)
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// Destructor

MyMatImpl::~MyMatImpl()
{
    Clear();
}

//////////////////////////////////////////////////////////////////////////////
// Clear

void MyMatImpl::Clear()
{
    m_numRows = 0;
    m_numColumns = 0;
    m_bSymmetric = false;

    m_columns.clear();

    delete m_preconditioner;   m_preconditioner = 0;
    delete m_values;           m_values = 0;
    delete m_colPtr;           m_colPtr = 0;
    delete m_rowInd;           m_rowInd = 0;
}

//////////////////////////////////////////////////////////////////////////////
// BeginConstruction - Sets the size and prepares for the construction phase

void MyMatImpl::BeginConstruction( int numRows, int numColumns, bool bSymmetric )
{
    Clear();

    if( numRows > 0 && numColumns > 0 )
    {
        m_numRows = numRows;
        m_numColumns = numColumns;
        m_columns.resize( numColumns );
    }

    m_bSymmetric = bSymmetric;
}

//////////////////////////////////////////////////////////////////////////////
// EndConstruction - Wraps up the construction phase. Entries cannot be
//                   modified after this function has been called

void MyMatImpl::EndConstruction( PRECOND_TYPE_ENUM type, real dropTolerance )
{
    // Convert to the compressed representation
    m_colPtr = new int[m_numColumns+1];
    int size = 0;
    for( int i = 0; i < m_numColumns; ++i )
    {
        m_colPtr[i] = size;
        size += (int)m_columns[i].size();
    }
    m_colPtr[m_numColumns] = size;

    m_values = new double[size];
    m_rowInd = new int[size];
    int j = 0;
    for( int i = 0; i < m_numColumns; ++i )
    {
        for( map<int,double>::const_iterator colIter = m_columns[i].begin();
            colIter != m_columns[i].end(); ++colIter, ++j )
        {
            m_rowInd[j] = colIter->first;
            m_values[j] = colIter->second;
        }
    }

    // Free up the memory from the other representation
    m_columns.clear();

    // Initialize the preconditioner
    if( m_bSymmetric )
    {
        switch( type )
        {
        case PRECOND_TYPE_DIAGONAL:
            m_preconditioner = new DiagonalPreconditioner( *this );
            break;
        case PRECOND_TYPE_INCOMPLETE_CHOLESKY:
            m_preconditioner = new IncompleteCholeskyPreconditioner( *this, dropTolerance );
            break;
        default:
            m_preconditioner = new IdentityPreconditioner();
            break;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
// Save - to file

void MyMatImpl::Save( const char *filename )
{
    ofstream out( filename, ios::out | ios::binary );

    out.write( (char*)(&m_numRows), sizeof(m_numRows) );
    out.write( (char*)(&m_numColumns), sizeof(m_numColumns) );
    out.write( (char*)(&m_bSymmetric), sizeof(m_bSymmetric) );

    out.write( (char*)(m_colPtr), (m_numColumns+1) * sizeof(int) );
    out.write( (char*)(m_values), m_colPtr[m_numColumns] * sizeof(double) );
    out.write( (char*)(m_rowInd), m_colPtr[m_numColumns] * sizeof(int) );

    out.close();
}

//////////////////////////////////////////////////////////////////////////////
// Load - from file

void MyMatImpl::Load( const char *filename )
{
    Clear();

    ifstream in( filename, ios::in | ios::binary );

    in.read( (char*)(&m_numRows), sizeof(m_numRows) );
    in.read( (char*)(&m_numColumns), sizeof(m_numColumns) );
    in.read( (char*)(&m_bSymmetric), sizeof(m_bSymmetric) );

    m_colPtr = new int[m_numColumns+1];
    in.read( (char*)(m_colPtr), (m_numColumns+1) * sizeof(int) );

    m_values = new double[ m_colPtr[m_numColumns] ];
    m_rowInd = new int[ m_colPtr[m_numColumns] ];
    in.read( (char*)(m_values), m_colPtr[m_numColumns] * sizeof(double) );
    in.read( (char*)(m_rowInd), m_colPtr[m_numColumns] * sizeof(int) );

    in.close();

    // Initialize the preconditioner
    if( m_bSymmetric )
        m_preconditioner = new IncompleteCholeskyPreconditioner( *this, (real)1.0e-1 );
}

//////////////////////////////////////////////////////////////////////////////
// Set - Sets the value of an entry. Creates the entry if it doesn't exist,
//       overwrites the entry if it does exist.

void MyMatImpl::Set( int i, int j, real value )
{
    smart_assert( i < m_numRows && j < m_numColumns )(i)(j)(m_numRows)(m_numColumns).fatal();
    // if it's symmetric we only store the lower triangle
    if( m_bSymmetric && i < j )
        swap(i,j);
    m_columns[j][i] = value;
}

//////////////////////////////////////////////////////////////////////////////
// Add - Adds the value of an entry. If it doesn't exist, assumes it was 0

void MyMatImpl::Add( int i, int j, real value )
{
    smart_assert( i < m_numRows && j < m_numColumns )(i)(j)(m_numRows)(m_numColumns).fatal();
    // if it's symmetric we only store the lower triangle
    if( m_bSymmetric && i < j )
        swap(i,j);
    m_columns[j][i] += value;
}

//////////////////////////////////////////////////////////////////////////////
// Multiply - A Matrix-Vector multiply.  product := this * vec

void MyMatImpl::Multiply( const VectorImpl *vec, VectorImpl *product ) const
{
    //assert( vec.size() == m_numColumns && product.size() == m_numRows );

    // Why all the casts instead of using operator[] directly, you ask?
    // Well, it turns out that operator[] gets called so unbelievably much
    // inside the inner loop of the inner loop of the solver, that the
    // function call overhead alone was accounting for over 25% of the
    // running time of the entire app! And of course it can't be inlined
    // because it's virtual. So now I access the data member directly,
    // and the running time has been cut in HALF! (Literally!)

    product->Assign( m_numRows, 0.0f );

    int j = 0;
    if( m_bSymmetric )
    {
        for( int i = 0; i < m_numColumns; ++i )
        {
            // The diagonal is always the first entry in each column

            ((MyVectorImpl*)product)->data[i] += m_values[j++] * ((const MyVectorImpl*)vec)->data[i];
            //(*product)[i] += m_values[j++] * (*vec)[i];

            for( ; j < m_colPtr[i+1]; ++j )
            {
                ((MyVectorImpl*)product)->data[m_rowInd[j]] += m_values[j] * ((const MyVectorImpl*)vec)->data[i];
                ((MyVectorImpl*)product)->data[i] += m_values[j] * ((const MyVectorImpl*)vec)->data[m_rowInd[j]];
                //(*product)[m_rowInd[j]] += m_values[j] * (*vec)[i];
                //(*product)[i] += m_values[j] * (*vec)[m_rowInd[j]];
            }
        }
    }
    else
    {
        for( int i = 0; i < m_numColumns; ++i )
            for( ; j < m_colPtr[i+1]; ++j )
                ((MyVectorImpl*)product)->data[m_rowInd[j]] += m_values[j] * ((const MyVectorImpl*)vec)->data[i];
                //(*product)[m_rowInd[j]] += m_values[j] * (*vec)[i];
    }
}

//////////////////////////////////////////////////////////////////////////////
// MAD -  product := plus - this * vec
//       (WARNING! it's not exactly a "MAD", that's why I made the function private)

void MyMatImpl::MAD( const VectorImpl *vec, const VectorImpl *plus, VectorImpl *product ) const
{
    //assert( vec.size() == m_numColumns && plus.size() == m_numRows && product.size() == m_numRows );

    product->Copy( plus );

    int j = 0;
    if( m_bSymmetric )
    {
        for( int i = 0; i < m_numColumns; ++i )
        {
            // The diagonal is always the first entry in each column

            ((MyVectorImpl*)product)->data[i] -= m_values[j++] * ((const MyVectorImpl*)vec)->data[i];
            //(*product)[i] -= m_values[j++] * (*vec)[i];

            for( ; j < m_colPtr[i+1]; ++j )
            {
                ((MyVectorImpl*)product)->data[m_rowInd[j]] -= m_values[j] * ((const MyVectorImpl*)vec)->data[i];
                ((MyVectorImpl*)product)->data[i] -= m_values[j] * ((const MyVectorImpl*)vec)->data[m_rowInd[j]];

                //(*product)[m_rowInd[j]] -= m_values[j] * (*vec)[i];
                //(*product)[i] -= m_values[j] * (*vec)[m_rowInd[j]];
            }
        }
    }
    else
    {
        for( int i = 0; i < m_numColumns; ++i )
            for( ; j < m_colPtr[i+1]; ++j )
                ((MyVectorImpl*)product)->data[m_rowInd[j]] -= m_values[j] * ((const MyVectorImpl*)vec)->data[i];
                //(*product)[m_rowInd[j]] -= m_values[j] * (*vec)[i];
    }
}

//////////////////////////////////////////////////////////////////////////////
// Solve - A preconditioned Conjugate Gradient solver

bool MyMatImpl::Solve( const VectorImpl *b, VectorImpl *x, int maxIterations, real epsilon ) const
{
    //assert( m_bSymmetric && x.size() == m_numColumns && b.size() == m_numRows );

    MyVectorImpl d(m_numRows), r(m_numRows), q(m_numRows), s(m_numRows);
    real delta, deltaOld, alpha, beta;

    this->MAD( x, b, &r );
    m_preconditioner->Apply( &r, &d );

    delta = r.DotProduct( &d );

    real tolerance = epsilon * epsilon * delta;

    int i;
    for( i = 0; i < maxIterations && delta > tolerance; ++i )
    {
        this->Multiply( &d, &q );
        alpha = delta / d.DotProduct( &q );
        x->MAD( &d, alpha );

        if( (i+1) % 50 == 0 )
            this->MAD( x, b, &r );
        else
            r.MAD( &q, -alpha );

        m_preconditioner->Apply( &r, &s );
        deltaOld = delta;
        delta = r.DotProduct( &s );
        beta = delta / deltaOld;
        d.MAD( beta, &s );
    }

    if( delta > tolerance )
    {
        LOG(warn) << "CG solver failed to converge. Delta=" << delta << " after " << i << " iterations.";
        return false;
    }

    if( m_bLogConvergenceStats )
        LOG(info) << "CG solver terminated after " << i << " iterations, delta = " << delta;

    return true;
}
