// Heavily modified from Cass's original version
/*
    glh - is a platform-indepenedent C++ OpenGL helper library 


    Copyright (c) 2000 Cass Everitt
    Copyright (c) 2000 NVIDIA Corporation
    All rights reserved.

    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

     * Redistributions of source code must retain the above
       copyright notice, this list of conditions and the following
       disclaimer.

     * Redistributions in binary form must reproduce the above
       copyright notice, this list of conditions and the following
       disclaimer in the documentation and/or other materials
       provided with the distribution.

     * The names of contributors to this software may not be used
       to endorse or promote products derived from this software
       without specific prior written permission. 

       THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
       ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
       LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
       FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
       REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
       INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
       BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
       LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
       CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
       LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
       ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
       POSSIBILITY OF SUCH DAMAGE. 


    Cass Everitt - cass@@r3.nu
*/


#ifndef _LINEAR_H_
#define _LINEAR_H_

#include <cmath>
#include <cassert>

#ifdef _MSC_VER
#pragma warning( push )
#pragma warning( disable : 4201 )  // nameless struct/union nonstandard extension
#endif

#define     GLH_ZERO            T(0.0)
#define     GLH_ONE             T(1.0)
#define     GLH_TWO             T(2.0)
#define     GLH_EPSILON         T(1e-6)
#define     GLH_PI              T(3.1415926535897932384626433832795)


template <class T> inline equivalent(T a,T b)
{ return ( (a < b + GLH_EPSILON) && (a > b - GLH_EPSILON) ) ? true : false; }


template <int N, class T>
class vec
{
public:
    int size() const { return N; }

    vec(const T & t = T()) 
    { for(int i = 0; i < N; i++) v[i] = t; }
    vec(const T * tp)
    { for(int i = 0; i < N; i++) v[i] = tp[i]; }

    const T * get_value() const
    { return v; }


    T dot( const vec<N,T> & rhs ) const
    {
        T r = 0;
        for(int i = 0; i < N; i++) r += v[i]*rhs.v[i];
        return r;
    }

    T length() const
    {
        T r = 0;
        for(int i = 0; i < N; i++) r += v[i]*v[i]; 
        return T(sqrt(r));
    }

    T square_norm() const
    {
        T r = 0;
        for(int i = 0; i < N; i++) r += v[i]*v[i]; 
        return r;
    }

    void  negate()
    { for(int i = 0; i < N; i++) v[i] = -v[i]; }


    T normalize()
    {
        T sum(0);
        for(int i = 0; i < N; i++) 
            sum += v[i]*v[i];
        sum = T(sqrt(sum));
        if (sum > GLH_EPSILON)
            for(int i = 0; i < N; i++) 
                v[i] /= sum;
        return sum;
    }


    vec<N,T> & set_value( const T * rhs )
    { for(int i = 0; i < N; i++) v[i] = rhs[i]; return *this; }

    T & operator [] ( int i )
    { return v[i]; }

    const T & operator [] ( int i ) const
    { return v[i]; }

    vec<N,T> & operator *= ( T d )
    { for(int i = 0; i < N; i++) v[i] *= d; return *this;}

    vec<N,T> & operator *= ( const vec<N,T> & u )
    { for(int i = 0; i < N; i++) v[i] *= u[i]; return *this;}

    vec<N,T> & operator /= ( T d )
    { if(d == 0) return *this; for(int i = 0; i < N; i++) v[i] /= d; return *this;}

    vec<N,T> & operator += ( const vec<N,T> & u )
    { for(int i = 0; i < N; i++) v[i] += u.v[i]; return *this;}

    vec<N,T> & operator -= ( const vec<N,T> & u )
    { for(int i = 0; i < N; i++) v[i] -= u.v[i]; return *this;}


    vec<N,T> operator - () const
    { vec<N,T> rv = v; rv.negate(); return rv; }

    vec<N,T> operator + ( const vec<N,T> &v) const
    { vec<N,T> rt(*this); return rt += v; }

    vec<N,T> operator - ( const vec<N,T> &v) const
    { vec<N,T> rt(*this); return rt -= v; }

    vec<N,T> operator * ( T d) const
    { vec<N,T> rt(*this); return rt *= d; }


    protected:
    T v[N];
};



// vector friend operators

template <int N, class T> inline
vec<N,T> operator * ( const vec<N,T> & b, T d )
{ vec<N,T> rt(b); return rt *= d; }

template <int N, class T> inline
vec<N,T> operator * ( T d, const vec<N,T> & b )
{ return b*d; }

template <int N, class T> inline
vec<N,T> operator * ( const vec<N,T> & b, const vec<N,T> & d )
{ vec<N,T> rt(b); return rt *= d; }

template <int N, class T> inline
vec<N,T> operator / ( const vec<N,T> & b, T d )
{ vec<N,T> rt(b); return rt /= d; }

template <int N, class T> inline
vec<N,T> operator + ( const vec<N,T> & v1, const vec<N,T> & v2 )
{ vec<N,T> rt(v1); return rt += v2; }

template <int N, class T> inline
vec<N,T> operator - ( const vec<N,T> & v1, const vec<N,T> & v2 )
{ vec<N,T> rt(v1); return rt -= v2; }

template <int N, class T> inline
bool operator == ( const vec<N,T> & v1, const vec<N,T> & v2 )
{
    for(int i = 0; i < N; i++)
        if(v1.v[i] != v2.v[i])
            return false;
    return true;
}

template <int N, class T> inline
bool operator != ( const vec<N,T> & v1, const vec<N,T> & v2 )
{ return !(v1 == v2); }


typedef vec<3,unsigned char> vec3ub;
typedef vec<4,unsigned char> vec4ub;



template <class T>
class _vec2 : public vec<2,T>
{
public:
    _vec2(const T & t = T()) : vec<2,T>(t)
    {}
    _vec2(const vec<2,T> & t) : vec<2,T>(t)
    {}
    _vec2(const T * tp) : vec<2,T>(tp)
    {}

    _vec2(T x, T y )
    { v[0] = x; v[1] = y; }

    const T * get_value() const
    { return v; }

    void get_value(T & x, T & y) const
    { x = v[0]; y = v[1]; }

    T cross( const _vec2 &rhs ) const
    { return v[0]*rhs.v[1]-v[1]*rhs.v[0]; }

    _vec2 & set_value( const T & x, const T & y)
    { v[0] = x; v[1] = y; return *this; }
};


template <class T>
class _vec3 : public vec<3,T>
{
public:
    _vec3(const T & t = T()) : vec<3,T>(t)
    {}
    _vec3(const vec<3,T> & t) : vec<3,T>(t)
    {}
    _vec3(const T * tp) : vec<3,T>(tp)
    {}

    _vec3(T x, T y, T z)
    { v[0] = x; v[1] = y; v[2] = z; }

    const T * get_value() const
    { return v; }

    void get_value(T & x, T & y, T & z) const
    { x = v[0]; y = v[1]; z = v[2]; }

    _vec3 cross( const _vec3 &rhs ) const
    {
        _vec3 rt;
        rt.v[0] = v[1]*rhs.v[2]-v[2]*rhs.v[1];
        rt.v[1] = v[2]*rhs.v[0]-v[0]*rhs.v[2];
        rt.v[2] = v[0]*rhs.v[1]-v[1]*rhs.v[0];
        return rt;
    }

    _vec3 & set_value( const T & x, const T & y, const T & z)
    { v[0] = x; v[1] = y; v[2] = z; return *this; }
};


template <class T>
class _vec4 : public vec<4,T>
{
public:
    _vec4(const T & t = T()) : vec<4,T>(t)
    {}
    _vec4(const vec<4,T> & t) : vec<4,T>(t)
    {}

    _vec4(const vec<3,T> & t, T fourth)
    { v[0] = t.v[0]; v[1] = t.v[1]; v[2] = t.v[2]; v[3] = fourth; }

    _vec4(const T * tp) : vec<4,T>(tp)
    {}

    _vec4(T x, T y, T z, T w)
    { v[0] = x; v[1] = y; v[2] = z; v[3] = w; }

    const T * get_value() const
    { return v; }

    void get_value(T & x, T & y, T & z, T & w) const
    { x = v[0]; y = v[1]; z = v[2]; w = v[3]; }

    _vec4 & set_value( const T & x, const T & y, const T & z, const T & w)
    { v[0] = x; v[1] = y; v[2] = z; v[3] = w; return *this; }
};

template <class T>
inline _vec3<T> homogenize(const _vec4<T> & v)
{
    _vec3<T> rt;
    assert(v.v[3] != GLH_ZERO);
    rt[0] = v[0]/v[3];
    rt[1] = v[1]/v[3];
    rt[2] = v[2]/v[3];
    return rt;
}























template <class T>
class _matrix4
{
public:
    _matrix4() { make_identity(); }

    _matrix4( T r )
    { set_value(r); }

    _matrix4( const T * m )
    { set_value(m); }

    _matrix4( T a00, T a01, T a02, T a03,
              T a10, T a11, T a12, T a13,
              T a20, T a21, T a22, T a23,
              T a30, T a31, T a32, T a33 )
    {
        element(0,0) = a00;
        element(0,1) = a01;
        element(0,2) = a02;
        element(0,3) = a03;

        element(1,0) = a10;
        element(1,1) = a11;
        element(1,2) = a12;
        element(1,3) = a13;

        element(2,0) = a20;
        element(2,1) = a21;
        element(2,2) = a22;
        element(2,3) = a23;

        element(3,0) = a30;
        element(3,1) = a31;
        element(3,2) = a32;
        element(3,3) = a33;
    }


    void get_value( T * mp ) const
    {
        int c = 0;
        for(int j=0; j < 4; j++)
            for(int i=0; i < 4; i++)
                mp[c++] = element(i,j);
    }


    const T * get_value() const
    { return m; }

    void set_value( const T * mp)
    {
        int c = 0;
        for(int j=0; j < 4; j++)
            for(int i=0; i < 4; i++)
                element(i,j) = mp[c++];
    }

    void set_value( T r )
    {
        for(int i=0; i < 4; i++)
            for(int j=0; j < 4; j++)
                element(i,j) = r;
    }

    void make_identity()
    {
        element(0,0) = 1.0;
        element(0,1) = 0.0;
        element(0,2) = 0.0;
        element(0,3) = 0.0;

        element(1,0) = 0.0;
        element(1,1) = 1.0;
        element(1,2) = 0.0;
        element(1,3) = 0.0;

        element(2,0) = 0.0;
        element(2,1) = 0.0;
        element(2,2) = 1.0;
        element(2,3) = 0.0;

        element(3,0) = 0.0;
        element(3,1) = 0.0;
        element(3,2) = 0.0;
        element(3,3) = 1.0;
    }


    static _matrix4 identity()
    {
        return _matrix4( 1.0, 0.0, 0.0, 0.0,
                         0.0, 1.0, 0.0, 0.0,
                         0.0, 0.0, 1.0, 0.0,
                         0.0, 0.0, 0.0, 1.0  );
    }


    void set_scale( T s )
    {
        element(0,0) = s;
        element(1,1) = s;
        element(2,2) = s;
    }

    void set_scale( const _vec3<T> & s )
    {
        element(0,0) = s[0];
        element(1,1) = s[1];
        element(2,2) = s[2];
    }


    void set_translate( const _vec3<T> & t )
    {
        element(0,3) = t[0];
        element(1,3) = t[1];
        element(2,3) = t[2];
    }

    void set_row(int r, const _vec4<T> & t)
    {
        element(r,0) = t[0];
        element(r,1) = t[1];
        element(r,2) = t[2];
        element(r,3) = t[3];
    }

    void set_column(int c, const _vec4<T> & t)
    {
        element(0,c) = t[0];
        element(1,c) = t[1];
        element(2,c) = t[2];
        element(3,c) = t[3];
    }


    void get_row(int r, _vec4<T> & t) const
    {
        t[0] = element(r,0);
        t[1] = element(r,1);
        t[2] = element(r,2);
        t[3] = element(r,3);
    }

    _vec4<T> get_row(int r) const
    {
        _vec4<T> v; get_row(r, v);
        return v;
    }

    void get_column(int c, _vec4<T> & t) const
    {
        t[0] = element(0,c);
        t[1] = element(1,c);
        t[2] = element(2,c);
        t[3] = element(3,c);
    }

    _vec4<T> get_column(int c) const
    {
        _vec4<T> v; get_column(c, v);
        return v;
    }

    _matrix4 inverse() const
    {
        _matrix4 minv;

        T r1[8], r2[8], r3[8], r4[8];
        T *s[4], *tmprow;

        s[0] = &r1[0];
        s[1] = &r2[0];
        s[2] = &r3[0];
        s[3] = &r4[0];

        register int i,j,p,jj;
        for(i=0;i<4;i++)
        {
            for(j=0;j<4;j++)
            {
                s[i][j] = element(i,j);
                if(i==j) s[i][j+4] = 1.0;
                else     s[i][j+4] = 0.0;
            }
        }
        T scp[4];
        for(i=0;i<4;i++)
        {
            scp[i] = T(abs(s[i][0]));
            for(j=1;j<4;j++)
                if(T(abs(s[i][j])) > scp[i]) scp[i] = T(abs(s[i][j]));
            if(scp[i] == 0.0) return minv; // singular matrix!
        }

        int pivot_to;
        T scp_max;
        for(i=0;i<4;i++)
        {
            // select pivot row
            pivot_to = i;
            scp_max = T(abs(s[i][i]/scp[i]));
            // find out which row should be on top
            for(p=i+1;p<4;p++)
                if(T(abs(s[p][i]/scp[p])) > scp_max)
                { scp_max = T(abs(s[p][i]/scp[p])); pivot_to = p; }
                // Pivot if necessary
                if(pivot_to != i)
                {
                    tmprow = s[i];
                    s[i] = s[pivot_to];
                    s[pivot_to] = tmprow;
                    T tmpscp;
                    tmpscp = scp[i];
                    scp[i] = scp[pivot_to];
                    scp[pivot_to] = tmpscp;
                }

                T mji;
                // perform gaussian elimination
                for(j=i+1;j<4;j++)
                {
                    mji = s[j][i]/s[i][i];
                    s[j][i] = 0.0;
                    for(jj=i+1;jj<8;jj++)
                        s[j][jj] -= mji*s[i][jj];
                }
        }
        if(s[3][3] == 0.0) return minv; // singular matrix!

        //
        // Now we have an upper triangular matrix.
        //
        //  x x x x | y y y y
        //  0 x x x | y y y y 
        //  0 0 x x | y y y y
        //  0 0 0 x | y y y y
        //
        //  we'll back substitute to get the inverse
        //
        //  1 0 0 0 | z z z z
        //  0 1 0 0 | z z z z
        //  0 0 1 0 | z z z z
        //  0 0 0 1 | z z z z 
        //

        T mij;
        for(i=3;i>0;i--)
        {
            for(j=i-1;j > -1; j--)
            {
                mij = s[j][i]/s[i][i];
                for(jj=j+1;jj<8;jj++)
                    s[j][jj] -= mij*s[i][jj];
            }
        }

        for(i=0;i<4;i++)
            for(j=0;j<4;j++)
                minv(i,j) = s[i][j+4] / s[i][i];

        return minv;
    }


    _matrix4 transpose() const
    {
        _matrix4 mtrans;
        for(int i=0;i<4;i++)
            for(int j=0;j<4;j++)
                mtrans(i,j) = element(j,i);
        return mtrans;
    }

    _matrix4 & mult_right( const _matrix4 & b )
    {
        _matrix4 mt(*this);
        set_value(T(0));

        for(int i=0; i < 4; i++)
            for(int j=0; j < 4; j++)
                for(int c=0; c < 4; c++)
                    element(i,j) += mt(i,c) * b(c,j);
        return *this;
    }    

    _matrix4 & mult_left( const _matrix4 & b )
    {
        _matrix4 mt(*this);
        set_value(T(0));

        for(int i=0; i < 4; i++)
            for(int j=0; j < 4; j++)
                for(int c=0; c < 4; c++)
                    element(i,j) += b(i,c) * mt(c,j);
        return *this;
    }

    // dst = M * src
    void mult_matrix_vec( const _vec3<T> &src, _vec3<T> &dst ) const
    {
        T w = ( src[0] * element(3,0) +
                src[1] * element(3,1) + 
                src[2] * element(3,2) +
                element(3,3)          );

        assert(w != GLH_ZERO);

        dst[0]  = ( src[0] * element(0,0) +
                    src[1] * element(0,1) +
                    src[2] * element(0,2) +
                    element(0,3)          ) / w;
        dst[1]  = ( src[0] * element(1,0) +
                    src[1] * element(1,1) +
                    src[2] * element(1,2) +
                    element(1,3)          ) / w;
        dst[2]  = ( src[0] * element(2,0) +
                    src[1] * element(2,1) + 
                    src[2] * element(2,2) +
                    element(2,3)          ) / w;
    }

    void mult_matrix_vec( _vec3<T> & src_and_dst) const
    { mult_matrix_vec(_vec3<T>(src_and_dst), src_and_dst); }


    // dst = src * M
    void mult_vec_matrix( const _vec3<T> &src, _vec3<T> &dst ) const
    {
        T w = ( src[0] * element(0,3) +
                src[1] * element(1,3) +
                src[2] * element(2,3) +
                element(3,3)          );

        assert(w != GLH_ZERO);

        dst[0]  = ( src[0] * element(0,0) +
                    src[1] * element(1,0) + 
                    src[2] * element(2,0) + 
                    element(3,0)          ) / w;
        dst[1]  = ( src[0] * element(0,1) +
                    src[1] * element(1,1) +
                    src[2] * element(2,1) +
                    element(3,1)          ) / w;
        dst[2]  = ( src[0] * element(0,2) +
                    src[1] * element(1,2) +
                    src[2] * element(2,2) +
                    element(3,2)          ) / w;
    }


    void mult_vec_matrix( _vec3<T> & src_and_dst) const
    { mult_vec_matrix(_vec3<T>(src_and_dst), src_and_dst); }

    // dst = M * src
    void mult_matrix_vec( const _vec4<T> &src, _vec4<T> &dst ) const
    {
        dst[0]  = ( src[0] * element(0,0) +
                    src[1] * element(0,1) +
                    src[2] * element(0,2) +
                    src[3] * element(0,3) );
        dst[1]  = ( src[0] * element(1,0) +
                    src[1] * element(1,1) +
                    src[2] * element(1,2) +
                    src[3] * element(1,3) );
        dst[2]  = ( src[0] * element(2,0) +
                    src[1] * element(2,1) + 
                    src[2] * element(2,2) +
                    src[3] * element(2,3) );
        dst[3]  = ( src[0] * element(3,0) +
                    src[1] * element(3,1) + 
                    src[2] * element(3,2) +
                    src[3] * element(3,3) );
    }

    void mult_matrix_vec( _vec4<T> & src_and_dst) const
    { mult_matrix_vec(_vec4<T>(src_and_dst), src_and_dst); }


    // dst = src * M
    void mult_vec_matrix( const _vec4<T> &src, _vec4<T> &dst ) const
    {
        dst[0]  = ( src[0] * element(0,0) +
                    src[1] * element(1,0) + 
                    src[2] * element(2,0) + 
                    src[3] * element(3,0) );
        dst[1]  = ( src[0] * element(0,1) +
                    src[1] * element(1,1) +
                    src[2] * element(2,1) +
                    src[3] * element(3,1) );
        dst[2]  = ( src[0] * element(0,2) +
                    src[1] * element(1,2) +
                    src[2] * element(2,2) +
                    src[3] * element(3,2) );
        dst[3]  = ( src[0] * element(0,3) +
                    src[1] * element(1,3) +
                    src[2] * element(2,3) +
                    src[3] * element(3,3) );
    }


    void mult_vec_matrix( _vec4<T> & src_and_dst) const
    { mult_vec_matrix(_vec4<T>(src_and_dst), src_and_dst); }


    // dst = M * src
    void mult_matrix_dir( const _vec3<T> &src, _vec3<T> &dst ) const
    {
        dst[0]  = ( src[0] * element(0,0) +
                    src[1] * element(0,1) +
                    src[2] * element(0,2) ) ;
        dst[1]  = ( src[0] * element(1,0) +
                    src[1] * element(1,1) +
                    src[2] * element(1,2) ) ;
        dst[2]  = ( src[0] * element(2,0) +
                    src[1] * element(2,1) + 
                    src[2] * element(2,2) ) ;
    }


    void mult_matrix_dir( _vec3<T> & src_and_dst) const
    { mult_matrix_dir(_vec3<T>(src_and_dst), src_and_dst); }


    // dst = src * M
    void mult_dir_matrix( const _vec3<T> &src, _vec3<T> &dst ) const
    {
        dst[0]  = ( src[0] * element(0,0) +
                    src[1] * element(1,0) +
                    src[2] * element(2,0) ) ;
        dst[1]  = ( src[0] * element(0,1) +
                    src[1] * element(1,1) +
                    src[2] * element(2,1) ) ;
        dst[2]  = ( src[0] * element(0,2) +
                    src[1] * element(1,2) + 
                    src[2] * element(2,2) ) ;
    }


    void mult_dir_matrix( _vec3<T> & src_and_dst) const
    { mult_dir_matrix(_vec3<T>(src_and_dst), src_and_dst); }


    T & operator () (int row, int col)
    { return element(row,col); }

    const T & operator () (int row, int col) const
    { return element(row,col); }

    T & element (int row, int col)
    { return m[row | (col<<2)]; }

    const T & element (int row, int col) const
    { return m[row | (col<<2)]; }

    _matrix4 & operator *= ( const _matrix4 & mat )
    {
        mult_right( mat );
        return *this;
    }

    _matrix4 & operator *= ( const T & r )
    {
        for (int i = 0; i < 4; ++i)
        {
            element(0,i) *= r;
            element(1,i) *= r;
            element(2,i) *= r;
            element(3,i) *= r;
        }
        return *this;
    }

    _matrix4 & operator += ( const _matrix4 & mat )
    {
        for (int i = 0; i < 4; ++i)
        {
            element(0,i) += mat.element(0,i);
            element(1,i) += mat.element(1,i);
            element(2,i) += mat.element(2,i);
            element(3,i) += mat.element(3,i);
        }
        return *this;
    }

    friend _matrix4 operator * ( const _matrix4 & m1, const _matrix4 & m2 );
    friend bool operator == ( const _matrix4 & m1, const _matrix4 & m2 );
    friend bool operator != ( const _matrix4 & m1, const _matrix4 & m2 );

    protected:
    T m[16];
};

template <class T>
inline _matrix4<T> operator * ( const _matrix4<T> & m1, const _matrix4<T> & m2 )
{
    static _matrix4<T> product;

    product = m1;
    product.mult_right(m2);

    return product;
}

// I don't know why, but the linker needs me to make this one explicit
inline _matrix4<float> operator * ( const _matrix4<float> & m1, const _matrix4<float> & m2 )
{
    static _matrix4<float> product;

    product = m1;
    product.mult_right(m2);

    return product;
}

template <class T>
inline bool operator ==( const _matrix4<T> &m1, const _matrix4<T> &m2 )
{
    return ( 
        m1(0,0) == m2(0,0) &&
        m1(0,1) == m2(0,1) &&
        m1(0,2) == m2(0,2) &&
        m1(0,3) == m2(0,3) &&
        m1(1,0) == m2(1,0) &&
        m1(1,1) == m2(1,1) &&
        m1(1,2) == m2(1,2) &&
        m1(1,3) == m2(1,3) &&
        m1(2,0) == m2(2,0) &&
        m1(2,1) == m2(2,1) &&
        m1(2,2) == m2(2,2) &&
        m1(2,3) == m2(2,3) &&
        m1(3,0) == m2(3,0) &&
        m1(3,1) == m2(3,1) &&
        m1(3,2) == m2(3,2) &&
        m1(3,3) == m2(3,3) );
}

template <class T>
inline bool operator != ( const _matrix4<T> & m1, const _matrix4<T> & m2 )
{ return !( m1 == m2 ); }  













template <class T>
class _quaternion
{
public:

    _quaternion()
    {
        q[0]=q[1]=q[2]=0.0;
        q[3]=GLH_ONE;
    }

    _quaternion( const T v[4] )
    {
        set_value( v );
    }


    _quaternion( T q0, T q1, T q2, T q3 )
    {
        set_value( q0, q1, q2, q3 );
    }


    _quaternion( const _matrix4<T> & m )
    {
        set_value( m );
    }


    _quaternion( const _vec3<T> &axis, T radians )
    {
        set_value( axis, radians );
    }


    _quaternion( const _vec3<T> &rotateFrom, const _vec3<T> &rotateTo )
    {
        set_value( rotateFrom, rotateTo );
    }

    _quaternion( const _vec3<T> & from_look, const _vec3<T> & from_up,
        const _vec3<T> & to_look, const _vec3<T>& to_up)
    {
        set_value(from_look, from_up, to_look, to_up);
    }

    const T * get_value() const
    {
        return  &q[0];
    }

    void get_value( T &q0, T &q1, T &q2, T &q3 ) const
    {
        q0 = q[0];
        q1 = q[1];
        q2 = q[2];
        q3 = q[3];
    }

    _quaternion & set_value( T q0, T q1, T q2, T q3 )
    {
        q[0] = q0;
        q[1] = q1;
        q[2] = q2;
        q[3] = q3;
        counter = 0;
        return *this;
    }

    void get_value( _vec3<T> &axis, T &radians ) const
    {
        radians = T(acos( q[3] ) * GLH_TWO);
        if ( radians == GLH_ZERO )
            axis = _vec3<T>( 0.0, 0.0, 1.0 );
        else
        {
            axis[0] = q[0];
            axis[1] = q[1];
            axis[2] = q[2];
            axis.normalize();
        }
    }

    void get_value( _matrix4<T> & m ) const
    {
        T s, xs, ys, zs, wx, wy, wz, xx, xy, xz, yy, yz, zz;

        T norm = q[0] * q[0] + q[1] * q[1] + q[2] * q[2] + q[3] * q[3];

        s = (equivalent(norm,GLH_ZERO)) ? GLH_ZERO : ( GLH_TWO / norm );

        xs = q[0] * s;
        ys = q[1] * s;
        zs = q[2] * s;

        wx = q[3] * xs;
        wy = q[3] * ys;
        wz = q[3] * zs;

        xx = q[0] * xs;
        xy = q[0] * ys;
        xz = q[0] * zs;

        yy = q[1] * ys;
        yz = q[1] * zs;
        zz = q[2] * zs;

        m(0,0) = T ( GLH_ONE - ( yy + zz ));
        m(1,0) = T ( xy + wz );
        m(2,0) = T ( xz - wy );

        m(0,1) = T ( xy - wz );
        m(1,1) = T ( GLH_ONE - ( xx + zz ));
        m(2,1) = T ( yz + wx );

        m(0,2) = T ( xz + wy );
        m(1,2) = T ( yz - wx );
        m(2,2) = T ( GLH_ONE - ( xx + yy ));

        m(3,0) = m(3,1) = m(3,2) = m(0,3) = m(1,3) = m(2,3) = GLH_ZERO;
        m(3,3) = GLH_ONE;
    }

    _quaternion & set_value( const T * qp )
    {
        memcpy(q,qp,sizeof(T) * 4);

        counter = 0;
        return *this;
    }

    _quaternion & set_value( const _matrix4<T> & m )
    {
        T tr, s;
        int i, j, k;
        const int nxt[3] = { 1, 2, 0 };

        tr = m(0,0) + m(1,1) + m(2,2);

        if ( tr > GLH_ZERO )
        {
            s = T(sqrt( tr + m(3,3) ));
            q[3] = T ( s * 0.5 );
            s = T(0.5) / s;

            q[0] = T ( ( m(1,2) - m(2,1) ) * s );
            q[1] = T ( ( m(2,0) - m(0,2) ) * s );
            q[2] = T ( ( m(0,1) - m(1,0) ) * s );
        }
        else
        {
            i = 0;
            if ( m(1,1) > m(0,0) )
                i = 1;

            if ( m(2,2) > m(i,i) )
                i = 2;

            j = nxt[i];
            k = nxt[j];

            s = T(sqrt( ( m(i,j) - ( m(j,j) + m(k,k) )) + GLH_ONE ));

            q[i] = T ( s * 0.5 );
            s = T(0.5 / s);

            q[3] = T ( ( m(j,k) - m(k,j) ) * s );
            q[j] = T ( ( m(i,j) + m(j,i) ) * s );
            q[k] = T ( ( m(i,k) + m(k,i) ) * s );
        }

        counter = 0;
        return *this;
    }

    _quaternion & set_value( const _vec3<T> &axis, T theta )
    {
        T sqnorm = axis.square_norm();

        if (sqnorm <= GLH_EPSILON)
        {
            // axis too small.
            x = y = z = 0.0;
            w = GLH_ONE;
        } 
        else 
        {
            theta *= T(0.5);
            T sin_theta = T(sin(theta));

            if (!equivalent(sqnorm,GLH_ONE)) 
                sin_theta /= T(sqrt(sqnorm));
            x = sin_theta * axis[0];
            y = sin_theta * axis[1];
            z = sin_theta * axis[2];
            w = T(cos(theta));
        }
        return *this;
    }

    _quaternion & set_value( const _vec3<T> & rotateFrom, const _vec3<T> & rotateTo )
    {
        _vec3<T> p1, p2;
        T alpha;

        p1 = rotateFrom; 
        p1.normalize();
        p2 = rotateTo;  
        p2.normalize();

        alpha = p1.dot(p2);

        if(equivalent(alpha,GLH_ONE))
        { 
            *this = identity(); 
            return *this; 
        }

        // ensures that the anti-parallel case leads to a positive dot
        if(equivalent(alpha,-GLH_ONE))
        {
            _vec3<T> v;

            if(p1[0] != p1[1] || p1[0] != p1[2])
                v = _vec3<T>(p1[1], p1[2], p1[0]);
            else
                v = _vec3<T>(-p1[0], p1[1], p1[2]);

            v -= p1 * p1.dot(v);
            v.normalize();

            set_value(v, GLH_PI);
            return *this;
        }

        p1 = p1.cross(p2);  
        p1.normalize();
        set_value(p1,T(acos(alpha)));

        counter = 0;
        return *this;
    }

    _quaternion & set_value( const _vec3<T> & from_look, const _vec3<T> & from_up,
        const _vec3<T> & to_look, const _vec3<T> & to_up)
    {
        _quaternion r_look = _quaternion(from_look, to_look);

        _vec3<T> rotated_from_up(from_up);
        r_look.mult_vec(rotated_from_up);

        _quaternion r_twist = _quaternion(rotated_from_up, to_up);

        *this = r_twist;
        *this *= r_look;
        return *this;
    }

    _quaternion & operator *= ( const _quaternion & qr )
    {
        _quaternion ql(*this);

        w = ql.w * qr.w - ql.x * qr.x - ql.y * qr.y - ql.z * qr.z;
        x = ql.w * qr.x + ql.x * qr.w + ql.y * qr.z - ql.z * qr.y;
        y = ql.w * qr.y + ql.y * qr.w + ql.z * qr.x - ql.x * qr.z;
        z = ql.w * qr.z + ql.z * qr.w + ql.x * qr.y - ql.y * qr.x;

        counter = counter + qr.counter;
        counter++;
        counter_normalize();
        return *this;
    }

    void normalize()
    {
        T rnorm = GLH_ONE / T(sqrt(w * w + x * x + y * y + z * z));
        if (equivalent(rnorm, GLH_ZERO))
            return;
        x *= rnorm;
        y *= rnorm;
        z *= rnorm;
        w *= rnorm;
        counter = 0;
    }

    friend bool operator == ( const _quaternion & q1, const _quaternion & q2 );      

    friend bool operator != ( const _quaternion & q1, const _quaternion & q2 );

    friend _quaternion operator * ( const _quaternion & q1, const _quaternion & q2 );

    bool equals( const _quaternion & r, T tolerance = GLH_EPSILON ) const
    {
        T t;

        t = (
            (q[0]-r.q[0])*(q[0]-r.q[0]) +
            (q[1]-r.q[1])*(q[1]-r.q[1]) +
            (q[2]-r.q[2])*(q[2]-r.q[2]) +
            (q[3]-r.q[3])*(q[3]-r.q[3]) );
        if(t > tolerance) 
            return false;
        return 1;
    }

    _quaternion & conjugate()
    {
        q[0] = -q[0];
        q[1] = -q[1];
        q[2] = -q[2];
        return *this;
    }

    _quaternion & invert()
    {
        return conjugate();
    }

    _quaternion inverse() const
    {
        _quaternion r = *this;
        return r.invert();
    }

    //
    // Quaternion multiplication with cartesian vector
    // v' = q*v*q(star)
    //
    void mult_vec( const _vec3<T> &src, _vec3<T> &dst ) const
    {
        T v_coef = w * w - x * x - y * y - z * z;
        T u_coef = GLH_TWO * (src[0] * x + src[1] * y + src[2] * z);  
        T c_coef = GLH_TWO * w;

        dst[0] = v_coef * src[0] + u_coef * x + c_coef * (y * src[2] - z * src[1]);
        dst[1] = v_coef * src[1] + u_coef * y + c_coef * (z * src[0] - x * src[2]);
        dst[2] = v_coef * src[2] + u_coef * z + c_coef * (x * src[1] - y * src[0]);
    }

    void mult_vec( _vec3<T> & src_and_dst) const
    {
        mult_vec(_vec3<T>(src_and_dst), src_and_dst);
    }

    void scale_angle( T scaleFactor )
    {
        _vec3<T> axis;
        T radians;

        get_value(axis, radians);
        radians *= scaleFactor;
        set_value(axis, radians);
    }

    static _quaternion slerp( const _quaternion & p, const _quaternion & q, T alpha )
    {
        _quaternion r;

        T cos_omega = p.x * q.x + p.y * q.y + p.z * q.z + p.w * q.w;
        // if B is on opposite hemisphere from A, use -B instead

        int bflip;
        if ( ( bflip = (cos_omega < GLH_ZERO)) )
            cos_omega = -cos_omega;

        // complementary interpolation parameter
        T beta = GLH_ONE - alpha;     

        if(cos_omega <= GLH_ONE - GLH_EPSILON)
            return p;

        T omega = T(acos(cos_omega));
        T one_over_sin_omega = GLH_ONE / T(sin(omega));

        beta    = T(sin(omega*beta)  * one_over_sin_omega);
        alpha   = T(sin(omega*alpha) * one_over_sin_omega);

        if (bflip)
            alpha = -alpha;

        r.x = beta * p.q[0]+ alpha * q.q[0];
        r.y = beta * p.q[1]+ alpha * q.q[1];
        r.z = beta * p.q[2]+ alpha * q.q[2];
        r.w = beta * p.q[3]+ alpha * q.q[3];
        return r;
    }

    static _quaternion identity()
    {
        return _quaternion( 0.0, 0.0, 0.0, GLH_ONE );
    }

    T & operator []( int i )
    {
        assert(i < 4);
        return q[i];
    }

    const T & operator []( int i ) const
    {
        assert(i < 4);
        return q[i];
    }

protected:

    void counter_normalize()
    {
        if (counter > 0)
            normalize();
    }

    union 
    {
        struct
        {
            T q[4];
        };
        struct
        {
            T x;
            T y;
            T z;
            T w;
        };
    };

    // renormalization counter
    unsigned char counter;
};

template <class T>
inline bool operator == ( const _quaternion<T> & q1, const _quaternion<T> & q2 )
{
    return (equivalent(q1.x, q2.x) &&
        equivalent(q1.y, q2.y) &&
        equivalent(q1.z, q2.z) &&
        equivalent(q1.w, q2.w) );
}

template <class T>
inline bool operator != ( const _quaternion<T> & q1, const _quaternion<T> & q2 )
{ 
    return ! ( q1 == q2 ); 
}

template <class T>
inline _quaternion<T> operator * ( const _quaternion<T> & q1, const _quaternion<T> & q2 )
{
    _quaternion<T> r(q1); 
    r *= q2; 
    return r; 
}

// I don't know why, but the linker needs me to make this one explicit
inline _quaternion<float> operator * ( const _quaternion<float> & q1, const _quaternion<float> & q2 )
{
    _quaternion<float> r(q1); 
    r *= q2; 
    return r; 
}


typedef _vec2<float> vec2f;
typedef _vec3<float> vec3f;
typedef _vec4<float> vec4f;
typedef _quaternion<float> quaternionf;
typedef _matrix4<float> matrix4f;

typedef _vec2<double> vec2d;
typedef _vec3<double> vec3d;
typedef _vec4<double> vec4d;
typedef _quaternion<double> quaterniond;
typedef _matrix4<double> matrix4d;


#include "util/real.h"

typedef _vec2<real> vec2;
typedef _vec3<real> vec3;
typedef _vec4<real> vec4;
typedef _quaternion<real> quaternion;
typedef _matrix4<real> matrix4;


#ifdef _MSC_VER
#pragma warning( pop )
#endif

#endif  // _LINEAR_H_
